export type Counter = number;

export enum OpCode {
  // flow
  COMPARE = 'cmp',
  JUMP = 'jmp',
  CALL = 'call',
  STOP = 'stop',
  // stack
  PEEK = 'peek',
  PEEK_STACK = 'peekstack',
  POP = 'pop',
  POP_STACK = 'popstack',
  PUSH = 'push',
  PUSH_STACK = 'pushstack',
  // math
  ADD = 'add',
  SUB = 'sub',
  MUL = 'mul',
  DIV = 'div',
  EXP = 'exp',
  MOD = 'mod',
}

export enum MathOperator {
  ADD = '+',
  SUB = '-',
  MUL = '*',
  DIV = '/',
  EXP = '^',
  MOD = '%',
}

export enum CompareOperator {
  EQUAL = '==',
  NOT_EQUAL = '!=',
  GREATER_THAN = '>',
  LESS_THAN = '<',
  GREATER_EQUAL = '>=',
  LESS_EQUAL = '<=',
}

export interface Address {
  type: 'address';
  module: number;
  function: number;
  offset: number;
}

export interface AddressRegisters {
  PC: Address;
  RA: Address;
  EH: Address;
  IH: Address;
}

export interface Register {
  type: 'data';
  value: number;
}

export interface DataRegisters {
  R1: Register;
  R2: Register;
  R3: Register;
  R4: Register;
}

export type ProgramRegisters = AddressRegisters & DataRegisters;

export type RawRegisters = {[K in keyof ProgramRegisters]: number};

export type RegisterName = keyof ProgramRegisters;

// nodes
export interface Node {
  type: string;
}

export interface RegisterNode {
  type: 'register';
  register: RegisterName;
}

export interface ConstantNode {
  type: 'constant';
  constant: number;
}

export interface CommandNode {
  type: 'command';
  code: OpCode;
  first: ValueNode;
  second?: ValueNode;
  third?: ValueNode;
}

export interface AssignmentNode {
  type: 'assignment';
  op: MathOperator;
  first: RegisterNode;
  second: ValueNode;
  third: ValueNode;
}

export interface BranchNode {
  body: Array<AnyNode>;
}

export interface ConditionalNode extends BranchNode {
  type: 'conditional';
  op: CompareOperator;
  first: ValueNode;
  second: ValueNode;
}

export interface FunctionNode extends BranchNode {
  type: 'function';
}

export type ValueNode = RegisterNode | ConstantNode;
export type AnyNode = RegisterNode | ConstantNode | CommandNode | AssignmentNode | ConditionalNode | FunctionNode;
export type NodeType = AnyNode['type'];

export const MATH_OP_CODE: Record<MathOperator, OpCode> = {
  '%': OpCode.MOD,
  '*': OpCode.MUL,
  '+': OpCode.ADD,
  '-': OpCode.SUB,
  '/': OpCode.DIV,
  '^': OpCode.EXP,
};

export const COMPARE_OP_CODE: Record<CompareOperator, OpCode> = {
  '==': OpCode.COMPARE,
  '!=': OpCode.COMPARE,
  '<': OpCode.COMPARE,
  '<=': OpCode.COMPARE,
  '>': OpCode.COMPARE,
  '>=': OpCode.COMPARE,
};
