import {
  AssignmentNode,
  BranchNode,
  CommandNode,
  ConditionalNode,
  ConstantNode,
  FunctionNode,
  MATH_OP_CODE,
  NodeType,
  OpCode,
  RegisterNode,
} from '../ast.js';
import { Program } from '../module/index.js';
import { StepResult } from './opcode/base.js';
import { COMPARE_CODES } from './opcode/compare.js';
import { FLOW_CODES } from './opcode/flow.js';
import { MATH_CODES } from './opcode/math.js';
import { STACK_CODES } from './opcode/stack.js';

export const NODE_STEP: Record<NodeType, (inst: any, state: Program) => boolean> = {
  'assignment': stepAssignment,
  'command': stepCommand,
  'conditional': stepConditional,
  'constant': stepConstant,
  'function': stepFunction,
  'register': stepRegister,
};

export const CODE_STEP = {
  ...COMPARE_CODES,
  ...FLOW_CODES,
  ...MATH_CODES,
  ...STACK_CODES,
}

export function stepAssignment(instruction: AssignmentNode, state: Program) {
  const command: CommandNode = {
    ...instruction,
    type: 'command',
    code: MATH_OP_CODE[instruction.op],
  };
  const result = stepCommand(command, state);
  console.log('post', result, state.registers);
  return result;
}

export function stepCommand(instruction: CommandNode, state: Program) {
  const result = stepCommandCode(instruction, state);
  state.registers.PC.offset += result.step;
  console.log('post', result, state.registers);
  return result.stop;
}

export function stepCommandCode(instruction: CommandNode, state: Program): StepResult {
  const codeFn = CODE_STEP[instruction.code];
  if (codeFn) {
    return codeFn(instruction, state.registers, state.stack);
  } else {
    throw new Error('unknown opcode');
  }
}

export function stepConditional(instruction: ConditionalNode, state: Program): boolean {
  const command: CommandNode = {
    ...instruction,
    type: 'command',
    code: OpCode.COMPARE,
  };
  const result = stepCommandCode(command, state);
  state.registers.PC.offset += 1; // abstract conditionals are always a single node
  if (result.step === 1) {
    // step through body
    console.log('step body', instruction.body);
    state.registers.RA = { ...state.registers.PC };
    for (let i = 0; i < instruction.body.length; ++i) {
      if (step(state, instruction, state.registers.RA.offset) === true) {
        return true;
      }
      if (state.registers.PC.offset < state.registers.RA.offset) {
        // CALL to earlier code
        break;
      }
      if (state.registers.PC.offset > state.registers.RA.offset + instruction.body.length) {
        // CALL to later code
        break;
      }
    }
    // run off end without CALL/JUMP
    console.log('done body', state.registers);
    state.registers.PC = { ...state.registers.RA };
  } else if (result.step === 2) {
    console.log('skip body', state.registers);
  }

  return false;
}

export function stepFunction(_instruction: FunctionNode, _state: Program): boolean {
  throw new Error('not implemented');
}

export function stepRegister(instruction: RegisterNode, _state: Program) {
  console.log('register', instruction);
  return false;
}

export function stepConstant(instruction: ConstantNode, _state: Program) {
  console.log('constant', instruction);
  return false;
}

/**
 * Step the program and execute one command.
 */
export function step(state: Program, fn: BranchNode, pc_offset = 0): boolean {
  const instruction = fn.body[state.registers.PC.offset - pc_offset];
  console.log('step', instruction, state.registers);

  const nodeFn = NODE_STEP[instruction.type];
  if (nodeFn) {
    return nodeFn(instruction, state);
  } else {
    throw new Error('unknown node');
  }
}
